# GIS4WRF (https://doi.org/10.5281/zenodo.1288569)
# Copyright (c) 2018 D. Meyer and M. Riechert. Licensed under MIT.

from typing import Tuple, Callable
import os

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import ( 
    QHBoxLayout, QVBoxLayout, QLabel, QLineEdit, QToolButton, QFileDialog, QAction, QGroupBox,
    QCheckBox, QPushButton, QMessageBox
)

from qgis.core import QgsApplication
from qgis.gui import QgsOptionsWidgetFactory, QgsOptionsPageWidget

from gis4wrf.core import get_wps_dist_url, get_wrf_dist_url, download_and_extract_dist, WRF_WPS_DIST_VERSION
from gis4wrf.core.util import export
from gis4wrf.plugin.options import get_options
from gis4wrf.plugin.constants import PLUGIN_NAME, GIS4WRF_LOGO_PATH
from gis4wrf.plugin.ui.helpers import FormattedLabel, TaskThread, WaitDialog, reraise, wrap_error

@export
class OptionsFactory(QgsOptionsWidgetFactory):
    def __init__(self):
        super().__init__()
        self.setTitle(PLUGIN_NAME)

    def icon(self):
        return QIcon(GIS4WRF_LOGO_PATH)

    def createWidget(self, parent):
        return ConfigOptionsPage(parent)

class ConfigOptionsPage(QgsOptionsPageWidget):
    def __init__(self, parent):
        super().__init__(parent)

        self.options = get_options()

        self.vbox = QVBoxLayout()
        self.setLayout(self.vbox)

        self.working_dir, layout = self.create_folder_input(
            'Working directory', self.options.working_dir)
        self.vbox.addLayout(layout)

        self.mpi_enabled, self.wps_dir, self.wrf_dir, gbox = self.create_distribution_box()
        self.vbox.addWidget(gbox)

        self.rda_username, self.rda_password, gbox = self.create_rda_auth_input()
        self.vbox.addWidget(gbox)

        self.vbox.addStretch()

    def apply(self) -> None:
        ''' Called when the options dialog is accepted. '''
        self.options.working_dir = self.working_dir.text()
        self.options.mpi_enabled = self.mpi_enabled.isChecked()
        self.options.wrf_dir = self.wrf_dir.text()
        self.options.wps_dir = self.wps_dir.text()
        self.options.rda_username = self.rda_username.text()
        self.options.rda_password = self.rda_password.text()

    def create_distribution_box(self) -> Tuple[QLineEdit, QLineEdit, QGroupBox]:
        gbox = QGroupBox('WPS/WRF Distribution')
        vbox = QVBoxLayout()
        gbox.setLayout(vbox)

        text = """<html>GIS4WRF allows you to run WPS and WRF on your local system.
                  If you have an existing compilation of WPS and/or WRF then simply choose
                  the corresponding folders below. If you compiled with <code>dmpar</code> make sure
                  to tick the "MPI" checkbox. Alternatively, you can download pre-compiled
                  distributions by clicking on the buttons below. Note that currently pre-compiled
                  distributions are not available for all platforms and are compiled using basic
                  nesting support.
                  </html>"""
        label = FormattedLabel(text, align=True)
        label = QLabel(text)
        label.setWordWrap(True)
        label.setOpenExternalLinks(True)
        vbox.addWidget(label)

        hbox = QHBoxLayout()
        mpi_enabled = QCheckBox('MPI')
        mpi_enabled.setChecked(self.options.mpi_enabled)
        hbox.addWidget(mpi_enabled)
        vbox.addLayout(hbox)

        wps_dir, hbox = self.create_folder_input('WPS directory', self.options.wps_dir)
        vbox.addLayout(hbox)

        wrf_dir, hbox = self.create_folder_input('WRF directory', self.options.wrf_dir)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        download_wps = QPushButton('Download Pre-Compiled WPS Distribution...')
        download_wrf = QPushButton('Download Pre-Compiled WRF Distribution...')
        download_wps.clicked.connect(self.download_wps)
        download_wrf.clicked.connect(self.download_wrf)
        hbox.addWidget(download_wps)
        hbox.addWidget(download_wrf)
        hbox.addStretch()
        vbox.addLayout(hbox)

        return mpi_enabled, wps_dir, wrf_dir, gbox

    def create_folder_input(self, label: str, value: str) -> Tuple[QLineEdit, QHBoxLayout]:
        hbox = QHBoxLayout()

        field = QLineEdit(value)

        button = QToolButton()
        action = QAction(QgsApplication.getThemeIcon('/mActionFileOpen.svg'), 'Choose Folder')
        button.setDefaultAction(action)

        def on_button_triggered():
            folder = QFileDialog.getExistingDirectory(caption='Select ' + label, directory=self.options.distributions_dir)
            if not folder:
                return
            field.setText(folder)

        button.triggered.connect(on_button_triggered)

        hbox.addWidget(QLabel(label))
        hbox.addWidget(field)
        hbox.addWidget(button)

        return field, hbox

    def create_rda_auth_input(self) -> Tuple[QLineEdit, QLineEdit, QGroupBox]:
        username = QLineEdit(self.options.rda_username)
        password = QLineEdit(self.options.rda_password)
        password.setEchoMode(QLineEdit.Password)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel('Username: '))
        hbox.addWidget(username)
        hbox.addWidget(QLabel('Password: '))
        hbox.addWidget(password)

        gbox = QGroupBox("NCAR's Research Data Archive (RDA)")
        text = """<html>GIS4WRF allows you to download datasets from
                <a href="https://rda.ucar.edu/">NCAR's Reseach Data Archive (RDA)</a>
                through the use of its API. If you do not have an RDA account, you need to
                <a href="https://rda.ucar.edu/index.html?hash=data_user&amp;action=register">register for a Data Account</a> first.
                Once you have completed your registration and your account is live you can save your log-in information to download meteorological
                data from GIS4WRF > Datasets > Met.</html>"""
        label = FormattedLabel(text, align=True)
        label = QLabel(text)
        label.setWordWrap(True)
        label.setOpenExternalLinks(True)
        vbox = QVBoxLayout()
        vbox.addWidget(label)
        vbox.addLayout(hbox)
        gbox.setLayout(vbox)

        return username, password, gbox

    def download_wps(self) -> None:
        mpi = self.mpi_enabled.isChecked()
        url = wrap_error(self, lambda: get_wps_dist_url(mpi))
        if not url:
            return

        self.download_dist('WPS', mpi, url, on_success=lambda folder: self.wps_dir.setText(folder))

    def download_wrf(self) -> None:
        mpi = self.mpi_enabled.isChecked()
        url = wrap_error(self, lambda: get_wrf_dist_url(mpi))
        if not url:
            return

        self.download_dist('WRF', mpi, url, on_success=lambda folder: self.wrf_dir.setText(folder))

    def download_dist(self, name: str, mpi: bool, url: str, on_success: Callable[[str],None]) -> None:
        mpi = self.mpi_enabled.isChecked()
        # TODO check if mpiexec/mpirun etc is installed and if not prompt user to install it

        if not self.confirm_dist_download(name, mpi, url):
            return

        wait_dialog = WaitDialog(self, 'Downloading...')
                
        folder = self.get_dist_download_folder(name, mpi)
        thread = TaskThread(lambda: download_and_extract_dist(url, folder))
        thread.finished.connect(lambda: wait_dialog.accept())
        thread.succeeded.connect(lambda: on_success(folder))
        thread.failed.connect(reraise)
        thread.start()

    def confirm_dist_download(self, name: str, mpi: bool, url: str) -> bool:
        reply = QMessageBox.question(self, 'Confirm Download',
            'You are about to download a pre-compiled distribution of {} {} in {} mode.\n\nURL: {}'.format(
                name, WRF_WPS_DIST_VERSION, 'MPI' if mpi else 'non-MPI', url),
            QMessageBox.Ok, QMessageBox.Cancel)
        return reply == QMessageBox.Ok

    def get_dist_download_folder(self, name: str, mpi: bool) -> str:
        return os.path.join(self.options.distributions_dir, 
                            '{}-{}-{}'.format(name, WRF_WPS_DIST_VERSION, ('' if mpi else 'no') + 'mpi'))
